import os, pandas as pd

from catrxneng import confs
from catrxneng.utils import Influx


class Expt:

    def __init__(self, reactor_class, rate_model, step_class, mcat):
        self.reactor_class = reactor_class
        self.rate_model = rate_model
        self.step_class = step_class
        self.mcat = mcat
        self.steps = []

    def add_step(self, step_name, start, end, T, p0, ghsv):
        step = self.step_class(step_name, len(self.steps) + 1, start, end)
        step.attach_reactor(self.reactor_class, self.rate_model, T, p0, ghsv, self.mcat)
        self.steps.append(step)

    def simulate(self, dt_sec, std_dev):
        for step in self.steps:
            step.simulate(dt_sec, std_dev)
        dframes = [step.time_series_data for step in self.steps]
        self.time_series_data = pd.concat(dframes, ignore_index=True)

    def upload_data(self, bucket, measurement):
        influx = Influx(
            url=os.getenv("INFLUXDB_URL"),
            org=os.getenv("INFLUXDB_ORG"),
            bucket=bucket,
            measurement=measurement,
        )
        tags = [
            confs.FR100.tags[data_id]
            for data_id in self.time_series_data.columns
            if data_id in self.time_series_data.columns
        ]
        influx.upload_dataframe(dataframe=self.time_series_data, tags=tags)



from .basic_plot import BasicPlot

# import plotly.graph_objects as go, numpy as np


# def plot(config):
#     fig = go.Figure()
#     for key, value in config["traces"].items():
#         x = value.get("x")
#         y = value["y"]
#         if x is not None:
#             xmin = np.min(x)
#             xmax = np.max(x)
#         if isinstance(y, (float, int)):
#             x = [xmin, xmax]
#             y = [y, y]
#         trace = go.Scatter(
#             x=x,
#             y=y,
#             mode=value["mode"],
#             name=key,
#             text=value.get("hover_labels"),
#             hovertemplate="<b>%{text}</b><br>X: %{x}<br>Y: %{y}<extra></extra>",
#         )
#         fig.add_trace(trace)

#     xmin = config.get("xmin")
#     xmax = config.get("xmax")
#     ymin = config.get("ymin")
#     ymax = config.get("ymax")

#     fig.update_layout(
#         title=dict(text=f"<b>{config['title']}</b>", x=0.5),
#         xaxis_title=f"<b>{config['xlabel']}</b>",
#         yaxis_title=f"<b>{config['ylabel']}</b>",
#         width=650,
#         # height=400,
#         height=None,
#         margin=dict(t=50, b=110),
#         # margin=dict(t=50, b=50),
#         yaxis=dict(
#             range=[ymin, ymax],
#             showline=True,
#             linecolor="black",
#             linewidth=2,
#             mirror=True,
#             nticks=9,
#         ),
#         xaxis=dict(
#             range=[xmin, xmax],
#             showline=True,
#             linecolor="black",
#             linewidth=2,
#             mirror=True,
#         ),
#         annotations=config.get("info_text"),
#         plot_bgcolor="white",
#         paper_bgcolor="white",
#         # colorway=px.colors.qualitative.D3
#         # colorway=px.colors.qualitative.Set1
#         # colorway=px.colors.qualitative.Pastel
#         # colorway=px.colors.qualitative.Dark2
#     )

#     fig.show()



import pandas as pd, numpy as np

from .step import Step
from catrxneng import confs


class Co2FtsStep(Step):

    def __init__(self, step_name, step_num, start, end):
        super().__init__(step_name, step_num, start, end)

    def attach_reactor(self, reactor_class, rate_model, T, p0, ghsv, mcat):
        self.reactor = reactor_class(rate_model, T, p0, ghsv, mcat, "co2")

    def simulate(self, dt_sec, std_dev):
        df = pd.DataFrame()
        df["timestamp"] = np.arange(self.start.UET, self.end.UET + dt_sec, dt_sec)
        num_points = df["timestamp"].size
        df["step_name"] = self.step_name
        df["step_num"] = self.step_num
        self.reactor.solve()
        df["bed_temp"] = self.reactor.T.C * np.random.normal(
            loc=1, scale=std_dev, size=num_points
        )
        df["pressure"] = self.reactor.P.bar * np.random.normal(
            loc=1, scale=std_dev, size=num_points
        )
        df["n2_mfc"] = self.reactor.F0["inert"].smLmin * np.random.normal(

            loc=1, scale=std_dev, size=num_points
        )
        df["co2_mfc"] = self.reactor.F0["co2"].smLmin * np.random.normal(
            loc=1, scale=std_dev, size=num_points
        )
        df["h2_mfc"] = self.reactor.F0["h2"].smLmin * np.random.normal(
            loc=1, scale=std_dev, size=num_points
        )

        ch4_sel = 0.3
        olefins_sel = 0.4
        paraffins_sel = 1 - ch4_sel - olefins_sel
        F_hydrocarbons = (
            self.reactor.F0["co2"] - self.reactor.F["co2"][-1] - self.reactor.F["co"][-1]
        )

        self.reactor.F["ch4"] = F_hydrocarbons * ch4_sel
        self.reactor.F["c2h4"] = F_hydrocarbons * olefins_sel / 4 / 2
        self.reactor.F["c3h6"] = F_hydrocarbons * olefins_sel / 4 / 3
        self.reactor.F["c4h8"] = F_hydrocarbons * olefins_sel / 4 / 4
        self.reactor.F["c5h10"] = F_hydrocarbons * olefins_sel / 4 / 5
        self.reactor.F["c2h6"] = F_hydrocarbons * paraffins_sel / 4 / 2
        self.reactor.F["c3h8"] = F_hydrocarbons * paraffins_sel / 4 / 3
        self.reactor.F["c4h10"] = F_hydrocarbons * paraffins_sel / 4 / 4
        self.reactor.F["c5h12"] = F_hydrocarbons * paraffins_sel / 4 / 5

        self.reactor.y = self.reactor.F / self.reactor.Ft

        for component in self.reactor.y.keys:
            df[f"{component}_gc_conc"] = self.reactor.y[
                component
            ].pct * np.random.normal(loc=1, scale=std_dev, size=num_points)

        self.time_series_data = df

    # def upload_data(self, bucket, measurement):
    #     influx = Influx(
    #         url=os.getenv("INFLUXDB_URL"),
    #         org=os.getenv("INFLUXDB_ORG"),
    #         bucket=bucket,
    #         measurement=measurement,
    #     )
    #     tags = [
    #         confs.FR100.tags[data_id]
    #         for data_id in self.time_series_data.columns
    #         if data_id in self.time_series_data.columns
    #     ]
    #     influx.upload_dataframe(dataframe=self.time_series_data, tags=tags)



from catrxneng.utils import rate_const
from catrxneng.reactions import RWGS
from catrxneng.quantities import *


class Brubach2022:
    def __init__(self):
        self.components = ["co2", "h2", "co", "h2o", "inert"]
        self.Ea_rwgs = Energy(kJmol=115)
        self.Ea_ft = Energy(kJmol=67.8)
        self.a_rwgs = Unitless(si=16.3)
        self.a_ft = Unitless(si=9.07)
        self.b_ft = InversePressure(inv_bar=2.44)
        self.kref_rwgs = RateConstant(order=1.5, molhgcatbar=8.13e-2)
        self.kref_ft = RateConstant(order=2, molhgcatbar=6.39e-2)
        self.Tref = Temperature(C=300)

    def k_rwgs(self, T):
        return rate_const(T=T, Ea=self.Ea_rwgs, kref=self.kref_rwgs, Tref=self.Tref)

    def k_ft(self, T):
        return rate_const(T=T, Ea=self.Ea_ft, kref=self.kref_ft, Tref=self.Tref)

    def r_rwgs(self, T, p):
        fwd = p["co2"] * p["h2"] ** 0.5
        rev = p["co"] * p["h2o"] / (RWGS(T).Keq * p["h2"] ** 0.5)
        num = self.k_rwgs(T) * (fwd - rev)
        denom = (1 + self.a_rwgs * p["h2o"] / p["h2"]) ** 2
        rate = num / denom
        return ReactionRate(si=rate.si)

    def r_ft(self, T, p):
        num = self.k_ft(T) * p["h2"] * p["co"]
        denom = (1 + self.a_ft * p["h2o"] / p[1] + self.b_ft * p["co"]) ** 2
        rate = num / denom
        return ReactionRate(si=rate.si)

    @property
    def rate(self):
        return {
            "co2": lambda T, p: -self.r_rwgs(T, p),
            "h2": lambda T, p: -self.r_rwgs(T, p) - 2 * self.r_ft(T, p),
            "co": lambda T, p: self.r_rwgs(T, p) - self.r_ft(T, p),
            "h2o": lambda T, p: self.r_rwgs(T, p) + self.r_ft(T, p),
            "inert": lambda T, p: ReactionRate(si=0)
        }



from datetime import datetime, timezone
import pytz

time_format = "%Y-%m-%d_%H%M"
date_format = "%Y-%m-%d"


class Time:

    def __init__(self, **kwargs):
        if kwargs:
            setattr(self, list(kwargs.keys())[0], list(kwargs.values())[0])
        else:
            self.UET = int(datetime.now(timezone.utc).timestamp())

    @property
    def UTC(self):
        return datetime.fromtimestamp(self.UET, tz=pytz.utc)

    @UTC.setter
    def UTC(self, value):
        self.UET = int(value.timestamp())

    @property
    def ET(self):
        return self.UTC.astimezone(pytz.timezone("US/Eastern"))

    @ET.setter
    def ET(self, value):
        self.UET = int(value.timestamp())

    @property
    def iso(self):
        return datetime.fromtimestamp(self.UET, tz=timezone.utc).isoformat()

    @iso.setter
    def iso(self, value):
        dt = datetime.fromisoformat(value)
        self.UET = int(dt.timestamp())

    @property
    def UTC_str(self):
        return self.UTC.strftime("%Y-%m-%d_%H%M")

    @UTC_str.setter
    def UTC_str(self, value):
        tz = pytz.timezone("UTC")
        utc = datetime.strptime(value, "%Y-%m-%d_%H%M").astimezone(tz)
        self.UET = int(utc.timestamp())

    @property
    def ET_str(self):
        return self.ET.strftime("%Y-%m-%d_%H%M")

    @ET_str.setter
    def ET_str(self, value):
        tz = pytz.timezone("US/Eastern")
        et = datetime.strptime(value, "%Y-%m-%d_%H%M").astimezone(tz)
        self.UET = int(et.timestamp())

    @property
    def date_str(self):
        return self.ET.strftime("%Y-%m-%d")

    @date_str.setter
    def date_str(self, value):
        tz = pytz.timezone("US/Eastern")
        dt = datetime.strptime(value, "%Y-%m-%d")
        dt = tz.localize(dt)
        self.UET = int(dt.timestamp())

    @property
    def ET_disp(self):
        return self.ET.strftime("%Y-%m-%d %H:%M")

    def __sub__(self, other):
        if isinstance(other, Time):
            return Time(UET=self.UET - other.UET)
        if isinstance(other, (float, int)):
            return Time(UET=self.UET - other)
        raise ValueError("'other' must be Time, float, or int.")

    def __rsub__(self, other):
        if isinstance(other, Time):
            return Time(UET=other.UET - self.UET)
        raise ValueError("'other' must be Time.")

    def __add__(self, other):
        if isinstance(other, (float, int)):
            return Time(UET=int(self.UET + other))
        if isinstance(other, Time):
            return Time(UET=self.UET + other.UET)
        raise ValueError("'other' must be Time, float, or int.")

    def __radd__(self, other):
        if isinstance(other, Time):
            return Time(UET=other.UET + self.UET)
        raise ValueError("'other' must be Time.")

    def __gt__(self, other):
        if isinstance(other, Time):
            return self.UET > other.UET
        return TypeError("'other' must be an instance of Time.")

    def __rgt__(self, other):
        if isinstance(other, Time):
            return other.UET > self.UET
        return TypeError("'other' must be an instance of Time.")

    def __lt__(self, other):
        if isinstance(other, Time):
            return self.UET < other.UET
        return TypeError("'other' must be an instance of Time.")

    def __rlt__(self, other):
        if isinstance(other, Time):
            return other.UET < self.UET
        return TypeError("'other' must be an instance of Time.")



import numpy as np, pandas as pd

from .influx import Influx
from .time import Time


def get_keys(obj1, obj2):
    if getattr(obj1, "keys", None) == getattr(obj2, "keys", None):
        keys = getattr(obj1, "keys", None)
    if getattr(obj1, "keys", None) is not None:
        keys = obj1.keys
    if getattr(obj2, "keys", None) is not None:
        keys = obj2.keys
    try:
        if keys is None:
            return None
        else:
            return keys.copy()
    except NameError:
        raise ValueError("Quantities have mismatching keys.")


def rate_const(T, Ea, k0=None, kref=None, Tref=None):
    from ..quantities import R, RateConstant

    if kref and Tref:
        k0 = kref / (np.exp(-Ea / (R * Tref)))
        k0 = RateConstant(si=k0.si, order=kref.order)

    k = k0 * np.exp(-Ea / (R * T))
    return RateConstant(si=k.si, order=k0.order)


def to_float(value):
    if isinstance(value, (np.ndarray, pd.Series)):
        return value.astype(float)
    if isinstance(value, int):
        return float(value)
    return value


def plot_info_box(text):
    text = text.replace("\\n", "\n")
    text = text.strip("\n")
    text = text.replace("\n", "<br>")
    annotations = [
        dict(
            x=0.5,
            y=-0.35,
            xref="paper",
            yref="paper",
            text=text,
            showarrow=False,
            font=dict(size=12, style="italic", weight="bold"),
        )
    ]
    return annotations


def divide(x, y):
    if isinstance(y, (np.ndarray, pd.Series)):
        y_safe = y.copy()
        y_safe[y_safe == 0] = np.nan
    else:
        y_safe = np.nan if y == 0 else y

    return x / y_safe


# def get_conf(conf_name):
#     conf_name = conf_name + ".conf"
#     conf_dir = os.path.join(os.path.dirname(__file__), "..", "..", "confs")
#     file_path = os.path.join(conf_dir, conf_name)
#     if not os.path.exists(file_path):
#         raise FileNotFoundError(
#             f"Configuration file '{conf_name}' not found in 'conf' folder."
#         )
#     with open(file_path, "r") as file:
#         return json.load(file)



from .ammonia_synthesis import AmmoniaSynthesis
from .co2_fts import CO2FTS
from .co_methanation import COMethanation
from .fts import FTS
from .rwgs import RWGS
from .sabatier import Sabatier
from .wgs import WGS


import plotly.graph_objects as go, numpy as np
from catrxneng.utils import plot_info_box


class BasicPlot:

    def __init__(self, title=None):
        self.fig = go.Figure()
        self.title = title

    def add_trace(self, x, y, name=None, mode="lines", hover_label=None):
        if isinstance(y, (float, int)):
            if x is not None:
                xmin = np.min(x)
                xmax = np.max(x)
            x = [xmin, xmax]
            y = [y, y]
        trace = go.Scatter(
            x=x,
            y=y,
            mode=mode,
            name=name,
            text=hover_label,
            hovertemplate="<b>%{text}</b><br>X: %{x}<br>Y: %{y}<extra></extra>",
        )
        self.fig.add_trace(trace)

    def render(self, xlabel, ylabel, xrange=None, yrange=None, info_text=None):
        if xrange is None:
            xrange = [None, None]
        if yrange is None:
            yrange = [None, None]

        top = 50
        width = 650
        if info_text is None:
            bottom = 50
            height = 400
        else:
            bottom = 110
            height = None

        self.fig.update_layout(
            title=dict(text=f"<b>{self.title}</b>", x=0.5),
            xaxis_title=f"<b>{xlabel}</b>",
            yaxis_title=f"<b>{ylabel}</b>",
            width=width,
            height=height,
            margin=dict(t=top, b=bottom),
            yaxis=dict(
                range=yrange,
                showline=True,
                linecolor="black",
                linewidth=2,
                mirror=True,
                nticks=9,
            ),
            xaxis=dict(
                range=xrange,
                showline=True,
                linecolor="black",
                linewidth=2,
                mirror=True,
            ),
            annotations=plot_info_box(info_text),
            plot_bgcolor="white",
            paper_bgcolor="white",
        )
        self.fig.show()



import os, pandas as pd
from influxdb_client import InfluxDBClient, WritePrecision

from catrxneng.utils.time import Time


class Influx:

    def __init__(self, url, org, bucket, measurement, data_conf):
        self.url = url
        self.org = org
        self.bucket = bucket
        self.measurement = measurement
        self.data_conf = data_conf

    def upload_dataframe(self, dataframe, tags=None):
        if tags is None:
            tags = dataframe.columns
            tags.remove("timestamp")
        dataframe.set_index("timestamp", inplace=True)
        with InfluxDBClient(
            url=self.url, token=os.getenv("INFLUXDB_TOKEN"), org=self.org
        ) as client:
            with client.write_api() as write_api:
                write_api.write(
                    bucket=self.bucket,
                    record=dataframe,
                    data_frame_measurement_name=self.measurement,
                    data_frame_tag_columns=tags,
                    write_precision=WritePrecision.S,
                )

    def generate_query(self, start: Time, end: Time, dt_sec, tags):
        tag_string = ""
        for tag in list(tags.values()):
            tag_string = tag_string + 'r["_field"] == "' + tag + '" or '
        tag_string = tag_string[:-4]
        tag_string = tag_string + ")"

        query_start = (
            start.UTC.strftime("%Y-%m-%d") + "T" + start.UTC.strftime("%H:%M:%S") + "Z"
        )
        query_end = (
            end.UTC.strftime("%Y-%m-%d") + "T" + end.UTC.strftime("%H:%M:%S") + "Z"
        )

        self.query = """
        from(bucket: "BUCKET")
        |> range(start: START, stop: END)
        |> filter(fn: (r) => r["_measurement"] == "MEASUREMENT")
        |> filter(fn: (r) => TAGS
        |> aggregateWindow(every: SECONDSs, fn: last, createEmpty: false)
        |> map(fn: (r) => ({ r with _time: int(v: r._time) / 1000000000}))
        |> keep(columns: ["_time", "_field", "_value"])
        |> yield(name: "last")
        |> pivot(rowKey:["_time"], columnKey: ["_field"], valueColumn: "_value")
        """

        self.query = self.query.replace("BUCKET", self.bucket)
        self.query = self.query.replace("MEASUREMENT", self.measurement)
        self.query = self.query.replace("START", query_start)
        self.query = self.query.replace("END", query_end)
        self.query = self.query.replace("TAGS", tag_string)
        self.query = self.query.replace("SECONDS", str(int(dt_sec)))

    def request_data(self):
        timeout_min = 5
        with InfluxDBClient(
            url=self.url,
            token=os.getenv("INFLUXDB_TOKEN"),
            org=self.org,
            timeout=timeout_min * 60 * 1000,
        ) as client:
            try:
                self.raw_dataframes = client.query_api().query_data_frame(
                    query=self.query
                )
            except ValueError:
                self.query = self.query.split("|> pivot")[0]
                self.raw_dataframes = client.query_api().query_data_frame(
                    query=self.query
                )

    def format_data(self):
        self.data = pd.concat(self.raw_dataframes, ignore_index=True)
        self.data = self.data[["_time", "_field", "_value"]]
        self.data.rename(
            columns={"_time": "timestamp", "_value": "value"}, inplace=True
        )
        self.data_dict = {
            key: self.data[self.data["_field"] == value][["timestamp", "value"]]
            for key, value in self.tags.items()
        }
        for key, dataset in self.data_dict.items():
            try:
                dataset.loc[:, "value"] = dataset["value"] * self.conf[key].get(
                    "multiplier", 1
                )
                dataset.loc[:, "value"] = dataset["value"] + self.conf[key].get(
                    "add", 0
                )
            except TypeError:
                pass



from .quantity import Quantity
from catrxneng.utils import to_float


class Moles(Quantity):

    @property
    def mol(self):
        return self.si

    @mol.setter
    def mol(self, value):
        self.si = to_float(value)

    


data_conf = {
    "bed_temp": {"reactor_attr": "T", "units": "C", "tag": "TIC02"},
    "pressure": {"reactor_attr": "P", "units": "bar", "tag": "PIC01"},
    "n2_mfc": {"reactor_attr": "Ft0", "units": "smLmin", "tag": "FIC01"},
    "co2_mfc": {"reactor_attr": "Ft0", "units": "smLmin", "tag": "FIC02"},
    "h2_mfc": {"reactor_attr": "Ft0", "units": "smLmin", "tag": "FIC03"},
}


import numpy as np
from scipy.optimize import minimize

from catrxneng.quantities import *


class Reaction:
    def __init__(self, T, limiting_reactant):
        self.T = T
        self.limiting_reactant = limiting_reactant
        self.fug_coeff = Unitless(
            si=np.ones(len(self.components)), keys=list(self.components.keys())
        )
        self.update()

    def update(self):
        for comp in self.components.values():
            comp.T = self.T
        self.Hf = Energy(si=[comp.Hf.si for comp in self.components.values()])
        self.Sf = Entropy(si=[comp.Sf.si for comp in self.components.values()])
        self.dHr = np.sum(self.Hf * self.stoich_coeff)
        self.dSr = np.sum(self.Sf * self.stoich_coeff)
        self.dGr = self.dHr - self.dSr * self.T
        self.Keq = np.exp(-self.dGr / (R * self.T))

    def check_components(self, p0):
        if p0.keys != list(self.components.keys):
            raise ValueError("Partial pressure keys do not match reaction components.")

    def equilibrate(self, p0):
        self.check_components(p0)
        P = np.sum(p0)
        initial_total_moles = Moles(si=100)
        initial_molfrac = p0 / P
        initial_moles = initial_molfrac * initial_total_moles
        std_state_fugacity = Pressure(atm=np.ones(len(self.components)))

        def objective(extent):

            extent = Moles(si=extent)
            moles = initial_moles + extent * self.stoich_coeff
            total_moles = np.sum(moles)
            molfrac = moles / total_moles
            fugacity = molfrac * self.fug_coeff * P
            activity = fugacity / std_state_fugacity
            Ka = np.prod(activity**self.stoich_coeff)
            # Kx = np.prod(molfrac**self.stoich_coeff)
            # Kphi = np.prod(self.fug_coeff**self.stoich_coeff)
            # Kp = np.prod(P**self.stoich_coeff)
            # Kf0 = np.prod(std_state_fugacity**self.stoich_coeff)
            return ((Ka - self.Keq) ** 2).si * 1e5

        adj_init_mol_reactants = np.array(
            [
                mol / stoich_coeff
                for mol, stoich_coeff in zip(initial_moles.si, self.stoich_coeff.si)
                if stoich_coeff < 0
            ]
        )
        min_extent = 1e-5
        bounds = [(min_extent, np.min(-adj_init_mol_reactants) * (1 - min_extent))]
        initial_guess = -0.5 * initial_moles.si[0] / self.stoich_coeff.si[0]
        result = minimize(
            objective,
            initial_guess,
            bounds=bounds,
            options={"ftol": 1e-10},
        )
        if result.success:
            self.extent = Moles(si=result.x[0])
            moles = initial_moles + self.extent * self.stoich_coeff
            self.conversion = (
                initial_moles[self.limiting_reactant] - moles[self.limiting_reactant]
            ) / initial_moles[self.limiting_reactant]
            total_moles = Moles(si=np.sum(moles.si))
            self.molfrac = moles / total_moles
        else:
            raise ValueError("Optimization failed: " + result.message)



from .quantity import Quantity
from .energy import Energy
from .temperature import Temperature
from .entropy import Entropy
from .pressure import Pressure
from .unitless import Unitless
from .mass import Mass
from .ghsv import GHSV
from .moles import Moles
from .molar_flow_rate import MolarFlowRate
from .reaction_rate import ReactionRate
from .rate_constant import RateConstant
from .vol_flow_rate import VolumetricFlowRate
from .inv_pressure import InversePressure
from .inv_ghsv import InvGHSV

from .gas_constant import GasConstant
R = GasConstant()


data_conf = {
    "bed_temp": {"reactor_attr": "T", "units": "C", "tag": "TIC02"},
    "pressure": {"reactor_attr": "P", "units": "bar", "tag": "PIC01"},
    "n2_mfc": {"reactor_attr": "Ft0", "units": "smLmin", "tag": "FIC01"},
    "co2_mfc": {"reactor_attr": "Ft0", "units": "smLmin", "tag": "FIC02"},
    "h2_mfc": {"reactor_attr": "Ft0", "units": "smLmin", "tag": "FIC03"},
}
components = {
    "n2": {"mol_weight": 28, "type": "inert"},
    "co2": {"mol_weight": 44, "c_atoms": 1, "o_atoms": 2, "type": "reactant"},
    "h2": {"mol_weight": 2, "type": "reactant", "h_atoms": 2},
    "co": {
        "mol_weight": 28,
        "c_atoms": 1,
        "o_atoms": 1,
        "type": "product",
    },
    "ch4": {
        "mol_weight": 16,
        "c_atoms": 1,
        "h_atoms": 4,
        "type": "product",
    },
    "c2h4": {
        "mol_weight": 28,
        "c_atoms": 2,
        "h_atoms": 4,
        "type": "product",
    },
    "c2h6": {
        "mol_weight": 30,
        "c_atoms": 2,
        "h_atoms": 6,
        "type": "product",
    },
    "c3h6": {
        "mol_weight": 42,
        "c_atoms": 3,
        "h_atoms": 6,
        "type": "product",
    },
    "c3h8": {
        "mol_weight": 44,
        "c_atoms": 3,
        "h_atoms": 8,
        "type": "product",
    },
    "c4h8": {
        "mol_weight": 56,
        "c_atoms": 4,
        "h_atoms": 8,
        "type": "product",
    },
    "c4h10": {
        "mol_weight": 58,
        "c_atoms": 4,
        "h_atoms": 10,
        "type": "product",
    },
    "c5h10": {
        "mol_weight": 70,
        "c_atoms": 5,
        "h_atoms": 10,
        "color": "green",
    },
    "c5h12": {
        "mol_weight": 72,
        "c_atoms": 5,
        "h_atoms": 12,
        "type": "product",
    },
    "h2o": {"mol_weight": 18, "type": "product", "h_atoms": 2, "o_atoms": 1},
}



import numpy as np
from scipy.integrate import solve_ivp

from .. import utils
from ..quantities import *
from .reactor import Reactor


class PFR(Reactor):
    def __init__(self, rate_model, T, p0, ghsv, mcat, limiting_reactant):
        self.rate_model = rate_model
        self.T = T
        self.p0 = p0
        self.ghsv = ghsv
        self.mcat = mcat
        self.limiting_reactant = limiting_reactant
        self.Ft0 = ghsv * mcat
        self.P = np.sum(p0)
        self.y0 = utils.divide(p0, self.P)
        self.F0 = self.y0 * self.Ft0
        # self.F = self.F0
        # self.extent = Moles(si=0)
        # self.conversion = Unitless(si=0)
        self.check_components()

    def dFdw(self, w, F):
        F = MolarFlowRate(molh=F)
        Ft = np.sum(F)
        y = F / Ft
        p = y * self.P
        return np.array(
            [rate(self.T, p).molhgcat for rate in self.rate_model.rate.values()]
        )

    def solve(self):
        w_span = (0, self.mcat.g)
        w_eval = np.linspace(0, self.mcat.g, 1000)
        solution = solve_ivp(self.dFdw, w_span, self.F0.molh, t_eval=w_eval)
        self.w = Mass(g=solution.t)
        self.F = MolarFlowRate(molh=solution.y, keys=self.rate_model.components.copy())
        self.conversion = utils.divide(
            self.F0[self.limiting_reactant] - self.F[self.limiting_reactant],
            self.F0[self.limiting_reactant],
        )
        self.inv_ghsv = InvGHSV(si=self.w.si / self.Ft0.si)
        self.ghsv = GHSV(si=utils.divide(1, self.inv_ghsv.si))
        vol_flow_rate = self.Ft0.si * R.si * self.T.si / self.P.si
        self.vol_flow_rate = VolumetricFlowRate(si=vol_flow_rate)
        self.tau_mod = Quantity(si=self.w.g / self.vol_flow_rate.mLs)

    @property
    def Ft(self):
        return np.sum(self.F, axis=0)

    # def rate(self):
    #     p = self.y * self.P
    #     rates = []
    #     for ri in self.rate_model.rate:
    #         rates.append(ri(self.T, p))
    #     return rates



import os, json, time, sys, pandas as pd, numpy as np
from influxdb_client import InfluxDBClient, Point
from influxdb_client.client.write_api import SYNCHRONOUS

import catrxneng as cat
from catrxneng.quantities import *


class StepwiseExpt:
    def __init__(self, reactor):
        pass
    
    def generate_single_data_points(self):
        rate_model = cat.kinetic_models.brubach2022.Brubach2022()
        cond = {
            "T": Temperature(C=300),
            "p0": Pressure([4,12,0,0,4]),
            "ghsv": GHSV(smLhgcat=170000),
            "mcat": Mass(g=0.25)
        }
        pfr = cat.reactors.PFR(cond, rate_model)
        pfr.solve()
        self.single_data_points = {
            "bed_temp": cond["T"].C,
            "pressure": np.sum(cond["p0"]).bar,
            "co2_mfc": (pfr.Ft0 * pfr.y0[0]).smLmin,
            "h2_mfc": (pfr.Ft0 * pfr.y0[1]).smLmin,
            "n2_mfc": (pfr.Ft0 * pfr.y0[4]).smLmin,
            ""
        }

    def generate_simulated_data(self, value, err, start, end, dt="3T"):
        timestamps = pd.date_range(start=start, end=end, freq=dt)
        simulated_values = value + err * np.random.randn(len(timestamps))
        data = pd.DataFrame({"timestamp": timestamps, "value": simulated_values})
        return data

    def write_data(self, df, bucket, measurement):
        client = InfluxDBClient(url=self.url, token=self.token, org=self.org)
        write_api = client.write_api(write_options=SYNCHRONOUS)
        records = [
            {
                "measurement": measurement,
                "time": row["timestamp"],
                "fields": {col: row[col] for col in df.columns if col != "timestamp"}
            }
            for _, row in df.iterrows()
        ]
        write_api.write(bucket=bucket, record=records)
        write_api.close()
        client.close()



from .expt import Expt

class Co2FtsExpt(Expt):

    def __init__(self, reactors):
        self.reactors = reactors

    def simulate(self):
        self.steps = []
        for reactor in self.reactors:




from .reaction import Reaction
from .. import species
from ..quantities import Unitless


class AmmoniaSynthesis(Reaction):
    def __init__(self, T):
        self.components = {
            "N2": species.N2(T=T),
            "H2": species.H2(T=T),
            "NH3": species.NH3(T=T),
            "inert": species.Ar(T=T)
        }
        self.stoich_coeff = Unitless(si=[-0.5, -1.5, 1, 0])
        super().__init__(T)



from .reaction import Reaction
from ..quantities import *
from ..species import *


class RWGS(Reaction):
    def __init__(self, T, limiting_reactant="co2"):
        self.components = {
            "co2": CO2(T=T),
            "h2": H2(T=T),
            "co": CO(T=T),
            "h2o": H2O(T=T),
            "inert": Ar(T=T),
        }
        self.stoich_coeff = Unitless(
            si=[-1.0, -1.0, 1.0, 1.0, 0.0], keys=list(self.components.keys())
        )
        super().__init__(T, limiting_reactant)
